// This file is part of Blue UNO. Please see UNOMain.java or http://blueuno.udd.be/ for
// license information. 
// Copyright (C) 2007 Niklas Udd
//
// This class is used by the client to handle cards. It keeps track of the current card aswell
// as all cards the player has. This class is also responsible for painting all cards.

package be.udd.blueuno;

import javax.microedition.lcdui.*;
import java.util.Vector;

public class CardHandler {
	public static final int NO_CURRENT_CARD = -1;

	public static final int SWITCH_DIRECTION = 10;
	public static final int STOP = 11;
	public static final int PLUS_TWO = 12;
	public static final int PLUS_FOUR = 13;

	private int cardWidth;
	private int cardHeight;

	private int currentCard;
	private Vector myCards;
	private int position;

	public CardHandler() {
		currentCard = NO_CURRENT_CARD;
		myCards = new Vector();
		position = 0;
	}

	public void setDimensions(int width, int height) {
		cardWidth = (width - 17) / 4;
		cardHeight = height - 12;
	}

	public int getSelectedCard() {
		return ((Integer)myCards.elementAt(position)).intValue();
	}

	public void setCurrentCard(int card) {
		currentCard = card;
	}

	public void addCard(int card) {
		addCard(new Integer(card));
	}

	public void addCard(Integer card) {
		int position = 0;
		boolean loopBroken = false;
		for(int i = 0; i < myCards.size(); i++) {
			position = i;
			if(((Integer)myCards.elementAt(i)).intValue() > card.intValue()) {
				loopBroken = true;
				break;
			}
		}

		if(loopBroken)
			myCards.insertElementAt(card, position);
		else
			myCards.addElement(card);
	}

	public boolean removeCard(int card) {
		boolean success = false;

		for(int i = 0; i < myCards.size(); i++) {
			if(((Integer)myCards.elementAt(i)).intValue() == card) {
				myCards.removeElementAt(i);
				success = true;
				break;
			}
		}

		if(success) {
			if(position >= myCards.size())
				position--;
		}

		return success;
	}

	public void paintCurrentCard(Graphics g, int x, int y) {
		if(currentCard != NO_CURRENT_CARD)
			paintCard(g, x, y, currentCard, false, true);
	}

	public void paintMyCards(Graphics g, int x, int y) {
		int start = 0;
		int stop;

		if(position > 0)
			start = position - 1;
		
		stop = start + 3;
		if(stop >= myCards.size()) {
			stop = myCards.size() - 1;
			if(position == stop)
				start -= 2;
			if(position == (stop -1))
				start--;
			if(start < 0)
				start = 0;
		}

		x += cardWidth/2 + 2;
		for(int i = start; i <= stop; i++) {
			paintCard(g, x, y, ((Integer)myCards.elementAt(i)).intValue(), (position == i), false);

			x += cardWidth + 4;
		}
	}

	private void paintCard(Graphics g, int x, int y, int card, boolean marked, boolean colorSelected) {
		int type = getType(card);
		int color = getColorNumber(card);

		if(color < 4) {
			g.setColor(colorNumberToColor(color));
			g.fillRect((x - cardWidth/2), (y - cardHeight/2), cardWidth, cardHeight);
		} else {
			// Vlja frg-kort
			if(colorSelected) {
				g.setColor(colorNumberToColor(card-56));
				g.fillRect((x - cardWidth/2), (y - cardHeight/2), cardWidth, cardHeight);
			} else {
				g.setColor(0xff0000);
				g.fillRect((x - cardWidth/2), (y - cardHeight/2), (cardWidth/2), (cardHeight/2));
				g.setColor(0x0000ff);
				g.fillRect((x - cardWidth/2), y, (cardWidth/2), (cardHeight/2) + 1);
				g.setColor(0x00ff00);
				g.fillRect(x, (y - cardHeight/2), (cardWidth/2) + 1, (cardHeight/2));
				g.setColor(0xffff00);
				g.fillRect(x, y, (cardWidth/2) + 1, (cardHeight/2) + 1);
			}
		}

		g.setColor(0x000000);
		g.drawRect((x - cardWidth/2), (y - cardHeight/2), cardWidth, cardHeight);

		if(marked) {
			g.setColor(0xff0000);
			g.drawRect((x - cardWidth/2 - 2), (y - cardHeight/2 - 2), (cardWidth + 4), (cardHeight + 4));
		}

		if(color < 4) {
			g.setColor(0x000000);
			Font standardFont = g.getFont();
			Font cardFont = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE);
			g.setFont(cardFont);
			g.drawString(typeToText(getType(card)), x, (y - cardFont.getHeight()/2), (Graphics.TOP|Graphics.HCENTER));
			g.setFont(standardFont);
		}
	}

	public static int colorNumberToColor(int color) {
		if(color == 0)
			return 0xff0000;
		if(color == 1)
			return 0x0000ff;
		if(color == 2)
			return 0x00ff00;
		if(color == 3)
			return 0xffff00;
		return 0xffffff;
	}

	public static String typeToText(int type) {
		if(type < 10)
			return (new Integer(type)).toString();
		if(type == 10)
			return "<>";
		if(type == 11)
			return "X";
		if(type == 12)
			return "+2";
		if(type == 13)
			return "+4";
		return "?";
	}

	public static int getColorNumber(int card) {
		return card/14;
	}

	public static int getType(int card) {
		return card % 14;
	}

	public void moveLeft() {
		if(position > 0)
			position--;
	}

	public void moveRight() {
		if(position < (myCards.size() - 1))
			position++;
	}

	public int size() {
		return myCards.size();
	}

	public int getPoints() {
		int points = 0;
		for(int i = 0; i < myCards.size(); i++) {
			int card = ((Integer)myCards.elementAt(i)).intValue();
			if(card > 55) {
				points += 50;
			} else {
				int type = getType(card);
				if(type < 10) {
					// 0-9
					points += type;
				} else if(type < 13) {
					// 10-12
					points += 20;
				} else {
					// 13
					points += 50;
				}
			}
		}
		return points;
	}
}