// This file is part of Blue UNO. Please see UNOMain.java or http://blueuno.udd.be/ for
// license information. 
// Copyright (C) 2007 Niklas Udd
//
// This class handles the actual game.

package be.udd.blueuno;

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;

public class GameScreen extends GameCanvas implements IUNOState, IConnectorNotifier, CommandListener {
	public static int BACKGROUND_COLOR_ACTIVE = 0xffffff;
	public static int BACKGROUND_COLOR_INACTIVE = 0xcccccc;

	private UNOMain controller;
	private Language language;
	private IConnector connector;

	private Form form;
	private Command putCardCommand;
	private Command getCardCommand;
	private Command sayUNOCommand;
	private Command nextCommand;
	private Command exitCommand;

	private CardHandler cards;
	private ColorChooser colorChooser;
	private MessageHandler messageHandler;

	private String currentPlayer;
	private boolean myTurn;
	private boolean hasSaidUNO;
	private boolean hasPutCard;
	private int numGotCards;

	public GameScreen(UNOMain controller, Language language, IConnector connector) {
		super(false);

		this.controller = controller;
		this.language = language;
		this.connector = connector;

		initializeVariables();
		createCommands();

		connector.setNotifier(this);
	}

	private void initializeVariables() {
		cards = new CardHandler();
		colorChooser = new ColorChooser(language);
		messageHandler = new MessageHandler(this, language);

		currentPlayer = "";
		myTurn = false;
		hasSaidUNO = false;
		hasPutCard = false;
		numGotCards = 0;
	}

	private void createCommands() {
		String getKey = getKeyName(getKeyCode(GAME_A));
		String sayKey = getKeyName(getKeyCode(GAME_B));
		String nextKey = getKeyName(getKeyCode(GAME_C));

		putCardCommand = new Command(language.get("game_put"), Command.SCREEN, 1);
		getCardCommand = new Command(language.get("game_draw") + " (" + getKey + ")", Command.SCREEN, 2);
		sayUNOCommand = new Command(language.get("game_uno") + " (" + sayKey + ")", Command.SCREEN, 3);
		nextCommand = new Command(language.get("done") + " (" + nextKey + ")", Command.SCREEN, 4);
		exitCommand = new Command(language.get("exit"), Command.EXIT, 5);

		setCommandListener(this);
		addCommand(exitCommand);
	}

	public void commandAction(Command c, Displayable d) {
		if (c == exitCommand) {
			try {
				connector.send("exit");
			} catch(Exception e) {}
			controller.exit();
		} else if(c == putCardCommand) {
			putCurrentCard();
		} else if(c == nextCommand) {
			nextPlayer();
		} else if(c == sayUNOCommand) {
			sayUNO();
		} else if(c == getCardCommand) {
			getCard();
		}
	}

	public void keyPressed(int keyCode) {
		if(getGameAction(keyCode) == LEFT) {
			if(colorChooser.isActive()) {
				colorChooser.moveLeft();
			} else {
				cards.moveLeft();
			}
			repaint();
		} else if(getGameAction(keyCode) == RIGHT) {
			if(colorChooser.isActive()) {
				colorChooser.moveRight();
			} else {
				cards.moveRight();
			}
			repaint();
		} else if(myTurn) {
			if(getGameAction(keyCode) == FIRE) {
				putCurrentCard();
			} else if(getGameAction(keyCode) == GAME_A) {
				getCard();
			} else if(getGameAction(keyCode) == GAME_B) {
				sayUNO();
			} else if(getGameAction(keyCode) == GAME_C) {
				nextPlayer();
			}
		}
	}

	public void messageReceived() {
		try {
			if(connector.hasReceived()) {
				String message = connector.receive();
				if(message.startsWith("get")) {
					cards.addCard(Integer.parseInt(message.substring(4)));
				} else if(message.startsWith("currentplayer")) {
					currentPlayer = message.substring(14);
				} else if(message.startsWith("currentcard")) {
					cards.setCurrentCard(Integer.parseInt(message.substring(12)));
				} else if(message.startsWith("put")) {
					myTurn = true;
					hasSaidUNO = false;
					hasPutCard = false;
					numGotCards = 0;
					addCommand(putCardCommand);
					addCommand(getCardCommand);
					addCommand(sayUNOCommand);
					addCommand(nextCommand);
				} else if(message.startsWith("accepted")) {
					int card = Integer.parseInt(message.substring(9));
					handleAccepted(card);
				} else if(message.startsWith("next")) {
					myTurn = false;
					removeCommand(putCardCommand);
					removeCommand(getCardCommand);
					removeCommand(sayUNOCommand);
					removeCommand(nextCommand);
				} else if(message.startsWith("show")) {
					messageHandler.handleShowMessage(message.substring(5));
				}

				if(message.startsWith("lost")) {
					try {
						connector.send("points " + cards.getPoints());
					} catch(Exception e) {}
					controller.gameLost(message.substring(5));
				} else {
					// There is no point in repainting this screen if the game is lost
					repaint();
				}
			}
		} catch(Exception e) {}
	}

	private void putCurrentCard() {
		// When the user wants to play a card, the number of the current card is sent to the server
		// with the play command. If the server accepts the card, i.e. if it is ok to play it, 
		// the server replies with an accepted command.

		int currentCard = cards.getSelectedCard();

		if(currentCard > 55) {
			// If the selected card is a Choose Color card, a color must first be chosen
			if(!colorChooser.isActive()) {
				colorChooser.activate();
				repaint();
				return;
			} else {
				// This is an interesting piece of code. After the user has chosen a 
				// color, the card actually selected is first removed. Once that is done
				// a card with the wanted color is added.
				// (See the web page, blueuno.udd.be, for details on how cards are
				// represented interally.)
				cards.removeCard(currentCard);
				currentCard = 56 + colorChooser.getSelectedColor();
				cards.addCard(currentCard);
				colorChooser.inactivate();
			}
		}

		try {
			connector.send("put " + currentCard);
		} catch(Exception e) {
			messageHandler.setMessageByCode("game_error");
		}
		repaint();
	}

	private void handleAccepted(int card) {
		hasPutCard = true;
		cards.removeCard(card);

		if(cards.size() == 0)
			handleWin();
	}

	private void handleWin() {
		WonScreen newScreen = new WonScreen(controller, language, connector);
		connector.setNotifier(newScreen);
		try { connector.send("won"); } catch(Exception e) {}
		controller.gameWon(newScreen);
	}

	private void nextPlayer() {
		if(hasPutCard || numGotCards >= 3) {
			try {
				if((cards.size() == 1) && !hasSaidUNO) {
					connector.send("nextnouno");
				} else {
					connector.send("next");
				}
			} catch(Exception e) {
				messageHandler.setMessageByCode("game_error");
			}
		} else {
			messageHandler.setMessageByCode("game_not_put");
		}
		repaint();
	}

	private void sayUNO() {
		if(cards.size() == 1) {
			try {
				hasSaidUNO = true;
				connector.send("uno");
			} catch(Exception e) {
				messageHandler.setMessageByCode("game_error");
			}
		} else {
			messageHandler.setMessageByCode("game_not_uno");
		}
		repaint();
	}

	private void getCard() {
		if(!hasPutCard && numGotCards < 3) {
			numGotCards++;
			try {
				connector.send("get");
			} catch(Exception e) {
				messageHandler.setMessageByCode("game_error");
			}
		} else {
			if(hasPutCard) {
				messageHandler.setMessageByCode("game_already_put");
			} else {
				messageHandler.setMessageByCode("game_three_drawn");
			}
		}
		repaint();
	}


	public void paint(Graphics g) {
		int x = g.getClipX();
		int y = g.getClipY();
		int width = g.getClipWidth();
		int height = g.getClipHeight();

		// Divide the screen into different parts
		int height1 = g.getFont().getHeight() + 5;
		int height4 = height1;
		int height2 = (height - height1 - height4) / 2;
		int height3 = height2;

		int currentBackgroundColor = (myTurn?BACKGROUND_COLOR_ACTIVE:BACKGROUND_COLOR_INACTIVE);

		cards.setDimensions(width, height2);

		g.setColor(0xaabbcc);
		g.fillRect(x, y, width, height1);

		g.setColor(currentBackgroundColor);
		g.fillRect(x, (y + height1), width, height2);
		g.fillRect(x, (y + height1 + height2), width, height3);
		g.fillRect(x, (y + height1 + height2 + height3), width, height4);

		g.setColor(0x000000);
		g.drawLine(x, (y + height1), (x + width), (y + height1));
		g.setColor(0x555555);
		g.drawLine(x, (y + height1 + height2), (x + width), (y + height1 + height2));

		g.setColor(0x000000);
		g.drawString(currentPlayer, (x + width/2), (y + 2), (Graphics.TOP|Graphics.HCENTER));

		cards.paintCurrentCard(g, (x + width/2), (y + height1 + height2/2));
		cards.paintMyCards(g, x, (y + height1 + height2 + height3/2));

		messageHandler.paint(g,x,y,width,height1,height2,height3,currentBackgroundColor);
		colorChooser.paint(g,x,y,width,height1,height2,height3);
	}


	public Displayable getDisplayable() {
		return this;
	}
}