// This file is part of Blue UNO. Please see UNOMain.java or http://blueuno.udd.be/ for
// license information. 
// Copyright (C) 2007 Niklas Udd
//
// This class is the server class.

package be.udd.blueuno;

import javax.bluetooth.*;
import javax.microedition.io.*;
import java.util.*;

public class Server {
	public static final String UUID = "B1E0B1E0B1E0B1E0B1E0B1E0B1E0B1";
	public static final boolean FORWARD = true;
	public static final boolean BACKWARD = false;

	private UNOMain controller;
	private ServerStarter starter;
	private Vector clients;
	private Random random;

	boolean keepListening = true;
	private int currentPlayer = 0;
	private int currentCard;
	private int lastCardFromThisPlayer = CardHandler.NO_CURRENT_CARD;
	private boolean direction = FORWARD;
	private boolean skipNextPlayer = false;

	private int winner;
	private int points;
	private int numUsersLeftToSendPoints;

	public Server(UNOMain controller, ServerStarter starter) {
		this.controller = controller;
		this.starter = starter;
		clients = new Vector();
		random = new Random();
	}

	public void addLocalPlayer(LocalConnector local) {
		clients.addElement(local);
	}

	public void start() {
		try {
			LocalDevice localDevice = LocalDevice.getLocalDevice();
	
			localDevice.setDiscoverable(DiscoveryAgent.GIAC);
			L2CAPConnectionNotifier notifier = (L2CAPConnectionNotifier)Connector.open("btl2cap://localhost:"+UUID+";name=BlueUNO");

			starter.serverStarted();	

			while(keepListening) {
				L2CAPConnection connection = notifier.acceptAndOpen();
				if(keepListening) {
					ClientHandlerThread clientHandlerThread = new ClientHandlerThread(controller, starter, connection);
					clients.addElement(clientHandlerThread);
					clientHandlerThread.start();
				} else {
					BluetoothConnector connector = new BluetoothConnector(connection);					
					try { connector.send("closed"); } catch(Exception e) {}
					connector.close();
				}
			}
		} catch(Exception e) {
			starter.serverStartFailed();
		}
	}

	public void stopAccepting() {
		keepListening = false;
	}

	public void startPlaying() {
		cleanUpInactiveConnections();

		// Tell all players that the game can start
		try {
			broadcast("start");
		} catch(Exception e) {
			// It might be a good idea to actually do something here. Perhaps unlinking the
			// player from the game.
		}

		try { Thread.sleep(500); } catch(Exception e) {}

		// Hand out cards
		for(int j = 0; j < 7; j++) {
			try {
				for(int i = 0; i < clients.size(); i++) {
					IConnector connector = (IConnector)clients.elementAt(i);
					connector.send("get " + selectRandomCard());
				}
			} catch(Exception e) {}
		}

		// Select a card to start with
		currentCard = selectRandomCard();
		broadcastCurrentCard();

		// Tell the first player to play
		activatePlayer();
		receive();
	}


	private void cleanUpInactiveConnections() {
		boolean done = false;
		while(!done) {
			done = true;
			for(int i = 1; i < clients.size(); i++) {
				if(!((ClientHandlerThread)clients.elementAt(i)).isActive()) {
					clients.removeElementAt(i);
					done = false;
					break;
				}
			}
		}
	}

	private void receive() {
		boolean gameEnded = false;

		// Check all players to see if anyone has sent a message
		while(!gameEnded) {
			for(int i = 0; i < clients.size(); i++) {
				if (((IConnector)clients.elementAt(i)).hasReceived()) {
					handleMessage(i, ((IConnector)clients.elementAt(i)).receive());
				}
			}

			try { Thread.sleep(200); } catch(Exception e) {}
		}
	}

	private void handleMessage(int player, String message) {
		try {
			if(message.startsWith("put")) {
				handlePutMessage(player, message);
			} else if(message.equals("uno")) {
				handleUNOMessage(player);
			} else if(message.startsWith("next")) {
				handleNextMessage(player, message);
			} else if(message.equals("exit")) {
				handleExitMessage(player);
			} else if(message.equals("get")) {
				handleGetMessage(player);
			} else if(message.equals("won")) {
				handleWonMessage(player);
			} else if(message.startsWith("points")) {
				handlePointsMessage(player, message);
			}
		} catch(Exception e) {
			send(player, "show error");
		}
	}

	private void handlePutMessage(int player, String message) throws Exception {
		int card = Integer.parseInt(message.substring(4));
		if(player == currentPlayer) {
			int color = CardHandler.getColorNumber(card);
			int type = CardHandler.getType(card);
			int currentColor = CardHandler.getColorNumber(currentCard);
			int currentType = CardHandler.getType(currentCard);
			boolean cardAccepted = false;

			if(card > 55) {
				// Choose Color card
				if((lastCardFromThisPlayer == CardHandler.NO_CURRENT_CARD) ||
					(lastCardFromThisPlayer > 55))
					cardAccepted = true;
			} else {
				if((lastCardFromThisPlayer == CardHandler.NO_CURRENT_CARD) && 
					(currentCard > 55)) {
					currentColor = currentCard - 56;
					if(color == currentColor)
						cardAccepted = true;
				} else {
					// Check type
					if(currentType == type)
						cardAccepted = true;
	
					// Check color (if this is the first card player by this player)
					if(lastCardFromThisPlayer == CardHandler.NO_CURRENT_CARD) {
						if(currentColor == color)
							cardAccepted = true;
					}
				}
			}

			if(cardAccepted) {
				currentCard = card;
				lastCardFromThisPlayer = card;

				if(type == CardHandler.SWITCH_DIRECTION) {
					direction = !direction;
				} else if(type == CardHandler.STOP) {
					skipNextPlayer = true;
				} else if(type == CardHandler.PLUS_TWO) {
					skipNextPlayer = true;
					addCardsToNextPlayer(2);
				} else if(type == CardHandler.PLUS_FOUR) {
					skipNextPlayer = true;
					addCardsToNextPlayer(4);
				}

				send(player, "accepted " + card);
				broadcastCurrentCard();
			} else {
				send(player, "show illegal");
			}
		} else {
			send(player, "show notturn");
		}
	}

	private void handleUNOMessage(int player) {
		try { broadcast("show uno " + getUsername(player)); } catch(Exception e) {}
	}

	private void handleNextMessage(int player, String message) {
		send(player, "next");
		if(message.equals("nextnouno")) {
			// The player forgot to say UNO

			send(player, "get " + selectRandomCard());
			try { Thread.sleep(500); } catch(Exception e) {}
			send(player, "get " + selectRandomCard());
			try { Thread.sleep(500); } catch(Exception e) {}
			send(player, "get " + selectRandomCard());

			String name = getUsername(player);
			try { broadcast("show nouno " + name); } catch(Exception e) {}
		}
		nextPlayer();
		activatePlayer();
	}

	private void handleExitMessage(int player) {
		// Someone wants to leave the game

		// Tries to change the player twice
		if(player == currentPlayer) {
			nextPlayer();
			if(player == currentPlayer)
				nextPlayer();
		}

		String name = getUsername(player);
		clients.removeElementAt(player);

		try { broadcast("show leave " + name); } catch(Exception e) {}

		activatePlayer();
	}

	private void handleGetMessage(int player) {
		send(player, "get " + selectRandomCard());		
	}

	private void handleWonMessage(int player) {
		String name = getUsername(player);
		points = 0;
		numUsersLeftToSendPoints = 0;
		winner = player;

		// Tell all players to send in their points
		for(int i = 0; i < clients.size(); i++) {
			try {
				if(i == player)
					continue;

				send(i, "lost " + name);
				numUsersLeftToSendPoints++;
			} catch(Exception e) {}
		}

		if(numUsersLeftToSendPoints == 0) {
			allPointsCollected();
		}
	}

	private void handlePointsMessage(int player, String message) {
		String playerPointsString = message.substring(7);

		int playerPoints = 0;
		try {
			playerPoints = Integer.parseInt(playerPointsString);
		} catch(Exception e) {}

		points += playerPoints;
		numUsersLeftToSendPoints--;

		if(numUsersLeftToSendPoints == 0) {
			allPointsCollected();
		}

		try { ((IConnector)clients.elementAt(player)).close(); } catch(Exception e) {}
	}

	private void allPointsCollected() {
		try {
			send(winner, "points " + points);
		} catch(Exception e) {}
	}

	private int selectRandomCard() {
		// There is room for improvement here!
		int value = random.nextInt() % 60;
		if(value < 0)
			value *= -1;
		return value;
	}

	private void broadcastCurrentCard() {
		try {
			broadcast("currentcard " + currentCard);
		} catch(Exception e) {}
	}

	private void activatePlayer() {
		String name = getUsername(currentPlayer);
		try {
			broadcast("currentplayer " + name);
		} catch(Exception e) {}
		try {
			((IConnector)clients.elementAt(currentPlayer)).send("put");
		} catch(Exception e) {}
	}

	private int calculateNextPlayer(boolean ignoreSkipFlag) {
		int numPlayers = clients.size();
		int nextPlayer = currentPlayer;
		int rounds = 1;
		if(skipNextPlayer && !ignoreSkipFlag)
			rounds++;

		for(int i = 0; i < rounds; i++) {
			if(direction == FORWARD) {
				nextPlayer++;
				if(nextPlayer >= numPlayers)
					nextPlayer = 0;
			} else {
				nextPlayer--;
				if(nextPlayer < 0)
					nextPlayer = numPlayers - 1;
			}
		}

		return nextPlayer;
	}

	private void nextPlayer() {
		currentPlayer = calculateNextPlayer(false);
		lastCardFromThisPlayer = CardHandler.NO_CURRENT_CARD;
		skipNextPlayer = false;
	}

	private void addCardsToNextPlayer(int numCards) {
		int nextPlayer = calculateNextPlayer(true);
		try {
			broadcast("show get " + getUsername(nextPlayer) + " " + numCards);
		} catch(Exception e) {}

		for(int i = 0; i < numCards; i++)
			send(nextPlayer, "get " + selectRandomCard());
	}

	private void send(int player, String message) {
		try {
			((IConnector)clients.elementAt(player)).send(message);
		} catch(Exception e) {}
	}

	private void broadcast(String message) throws Exception {
		for(int i = 0; i < clients.size(); i++) {
			IConnector connector = (IConnector)clients.elementAt(i);
			connector.send(message);
		}
	}

	private String getUsername(int position) {
		Object client = clients.elementAt(position);
		if(client instanceof ClientHandlerThread)
			return ((ClientHandlerThread)clients.elementAt(position)).getUsername();
		else
			return controller.getName();
	}
}